# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from uiPeer import *

class Viewer(UIPeer):
    Refresh = Pillow.In
    RefreshDone = Pillow.InState
    Document = Pillow.InState

    OUTER = 1
    INNER = 2 
    BOTH  = 3
    EXACT  = 4

    def __init__(self, room, cb, path, raw=False, refreshDoneCallback=None, exact=False, ignoreCredentials=False):
        UIPeer.__init__(self, room, path)
        self._valueCB = cb
        self._document = None

        self._exact = exact
        self._ignoreCredentials = ignoreCredentials
        self._raw = raw

        self._refreshDoneCallback = refreshDoneCallback

        self._catch(Viewer.In.Refresh, self._refresh)
        self._catch(Viewer.In.RefreshDone, self._refreshDone)
        self._catch(Viewer.In.Document, self._setDocument)

    @staticmethod
    def matchPath(path, changedPath, mode=3):
        if changedPath == None or path == None: return True

        if mode == Viewer.EXACT:
           if "*" in path:
                leftTokens  = path.split(".")
                rightTokens = changedPath.split(".")

                if len(rightTokens) == len(leftTokens): 
                    match = True

                    for i in range(min(len(leftTokens), len(rightTokens))):
                        if leftTokens[i] == "*": continue
                        if leftTokens[i] != rightTokens[i]: match = False

                    if match: return True
           elif path == changedPath: return True
           return False

        if mode in (Viewer.OUTER, Viewer.BOTH):
            if "*" in path:
                leftTokens  = path.split(".")
                rightTokens = changedPath.split(".")

                if len(rightTokens) <= len(leftTokens): 
                    match = True

                    for i in range(min(len(leftTokens), len(rightTokens))):
                        if leftTokens[i] == "*": continue
                        if leftTokens[i] != rightTokens[i]: match = False

                    if match: return True
            elif path.startswith(changedPath): return True

        if mode in (Viewer.INNER, Viewer.BOTH):
            if "*" in path:
                leftTokens  = path.split(".")
                rightTokens = changedPath.split(".")

                if len(leftTokens) <= len(rightTokens): 
                    match = True
    
                    for i in range(min(len(leftTokens), len(rightTokens))):
                        if leftTokens[i] == "*": continue
                        if leftTokens[i] != rightTokens[i]: match = False

                    if match: return True
            elif changedPath.startswith(path): return True

        return False

    def _cb(self, *args, **ka):
        if self._valueCB != None:
            self._valueCB(*args, **ka)

    def _refreshDone(self, pillow, path):
        if self._refreshDoneCallback:
            self._refreshDoneCallback()

    def selection(self):
        try:
            if self._document is not None: return self._document.getSelection(self._path)
        except: pass
        return None

    def _refresh(self, pillow, path):
        from wallaby.common.pathHelper import SelectionNotSpecified

        if self.destroyed: return

        if self._path == None:
            return

        if path == None:
            if self._raw:
                self._cb(None)
            else:
                self._cb('')
            return

        if self._exact and not Viewer.matchPath(self._path, path, mode=Viewer.EXACT): return
        elif self._raw and not Viewer.matchPath(self._path, path): return
        elif not self._raw and not Viewer.matchPath(self._path, path, mode=Viewer.OUTER): return

        try:
            value = None
            if self._document:
                value = self._document.get(self._path)


            if value != None and self._credentials:
                credential = self._credentials.getFlat(self._path)
                if self._path == '_id' or credential == None or 'view' in credential:
                    if self._raw:
                        self._cb(value)
                    else:
                        uvalue = unicode(value)
                        self._cb(uvalue)
            elif value != None and self._ignoreCredentials:
                if self._raw:
                    self._cb(value)
                else:
                    uvalue = unicode(value)
                    self._cb(uvalue)
            else:
                if self._raw:
                    self._cb(None)
                else:
                    self._cb('')
        except SelectionNotSpecified as e:
            if self._raw:
                self._cb(None)
            else:
                self._cb('-')
        except Exception as e:
            if self._raw:
                self._cb(None)
            else:
                self._cb('Invalid value')

    def _setDocument(self, pillow, document):
        if self.destroyed: return

        self._document = document
        self._refresh(pillow, self._path)

    def document(self):
        return self._document



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class TriggeredPillows(Peer):
    def __init__(self, room, pillows):
        Peer.__init__(self, room)

        self._triggers = {}

        for config in pillows:
            if config == None: continue
            trigger  = config.get("trigger", None)
            args     = config.get("trigger-args", None)
            pillow   = config.get("pillow", None)
            feathers = config.get("feathers", None)
            isJson   = config.get("isJson", False)

            if None in (trigger, pillow): continue

            if trigger not in self._triggers:
                self._triggers[trigger] = []

            self._triggers[trigger].append({
                "args": args,
                "pillow": pillow,
                "feathers": feathers,
                "isJson": isJson
            })

    def __translate(self, feathers, doc, path, value):
        if feathers != None:
            from wallaby.common.pathHelper import PathHelper
            import re
            if isinstance(feathers, (str, unicode)):
                if '%VALUE' in feathers:
                    match = re.match('(.*?)%VALUE:?(.*?)%(.*)', feathers)
                    if match.group(2):
                        try:
                            if match.group(1) or match.group(3):
                                if doc is not None:
                                    feathers = match.group(1) + unicode(doc.get(".".join( (path, match.group(2)) ))) + match.group(3)
                                else:
                                    feathers = match.group(1) + unicode(PathHelper.getValue(value, match.group(2))) + match.group(3)
                            else:
                                if doc is not None:
                                    feathers = doc.get(".".join( (path, match.group(2)) ))
                                else:
                                    feathers = PathHelper.getValue(value, match.group(2))
                        except:
                            feathers = None
            
                    else:
                        if match.group(1) or match.group(3):
                            feathers = match.group(1) + unicode(value) + match.group(3)
                        else:
                            feathers = value

                if '%path%' in feathers:
                    feathers = re.sub('%path%', path, feathers)

            if isinstance(feathers, dict):
                import copy
                feathers = copy.deepcopy(feathers)
                for k, v in feathers.items():
                    if isinstance(v, (str, unicode)) and '%VALUE' in v:
                        match = re.match('(.*?)%VALUE:?(.*?)%(.*)', v)

                        if match.group(2):
                            try:
                                if match.group(1) or match.group(3):
                                    if doc is not None:
                                        v = match.group(1) + unicode(doc.get(".".join( (path, match.group(2)) ))) + match.group(3)
                                    else:
                                        v = match.group(1) + unicode(PathHelper.getValue(value, match.group(2))) + match.group(3)
                                else:
                                    if doc is not None:
                                        v = doc.get(".".join( (path, match.group(2)) ))
                                    else:
                                        v = PathHelper.getValue(value, match.group(2))
                            except Exception as e:
                                v = None

                        else:
                            if match.group(1) or match.group(3):
                                v = match.group(1) + unicode(value) + match.group(3)
                            else:
                                v = value

                    if isinstance(v, (str, unicode)) and '%PATH%' in v:
                        v = re.sub('%PATH%', path, v)

                    feathers[k] = v
            
        return feathers 

    def trigger(self, trigger, args=None, doc=None, path=None, value=None, *ignore):
        if trigger not in self._triggers: 
            return False

        found = False

        for t in self._triggers[trigger]:    
            if args != None and args != t['args']: continue
            found = True

            feathers = t['feathers']            

            if t['isJson']:
                jsonString = '{"object":' + feathers + '}'
                try: 
                    import json
                    feathers = json.loads(jsonString)
                    feathers = feathers["object"]
                except Exception as e:
                    print "Error while parsing json", jsonString, e
                    feathers = None

            feathers = self.__translate(feathers, doc, path, value)
          
            if isinstance(t['args'], (str, unicode)):
                import re 
                match = re.match('^delay:(.*)$', t['args'])
                if match is not None:
                    try:
                        delay = float(match.group(1))
                        from twisted.internet import reactor
                        reactor.callLater(delay, self._throw, t['pillow'], feathers)
                        return
                    except: pass

            self._throw(t['pillow'], feathers)

        return found



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class RoomDebuggerParty(Peer):
    def __init__(self, room):
        Peer.__init__(self, room)

        from roomDebuggerRouting import RoomDebuggerRouting
        from credentialsRouting import CredentialsRouting
        from roomDebugger import RoomDebugger

        #Peers
        self.add(RoomDebuggerRouting(room))
        self.add(CredentialsRouting(room))
        self.add(RoomDebugger(room))



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

import wallaby.FX as FX
from twisted.internet import defer
from twisted.internet import task

class QueryResult(object):
    def __init__(self, source, query):
        self.query = query
        self._source = source

    def __str__(self):
        return "<"+self.__class__.__name__+"("+hex(id(self))+") with "+str(self.length())+" entries>"

    def source(self):
        return self._source

    def identifier(self):
        if self.query == None:
            return None
        else:
            return self.query.get('identifier')

    def deferredGetValue(self, index, path):
        return None

    def reorder(self, row, i1, i2):
        return None

    def deferredGetDocument(self, index):
        return None

    def deferredGetValues(self, indexList, path):
        values = []
        for index in indexList:
            values.append(self.deferredGetValue(index, path))
        return values

    def getDocumentID(self, index):
        return None

    def getDocumentIDs(self, indexList):
        documentIDs = []
        for index in indexList:
            documentIDs.append(self.getDocumentID(index))
        return documentIDs

    def length(self):
        return -1

    @FX.Property                                                                                                                                      
    def query():
        doc = "The query Property."
        def fget(self):
            return self._query
        def fset(self, value):
            self._query = value




# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class DocumentDatabase(Peer):
    Description = "Routing between document cache and database"

    from documentCache import DocumentCache
    from database import Database

    Routings = [
        (DocumentCache.Out.RequestDocument, Database.In.RequestDocument),
        (DocumentCache.Out.SaveDocument, Database.In.SaveDocument),
        (DocumentCache.Out.CreateDocument, Database.In.CreateDocument),
        (DocumentCache.Out.CreateAndSaveDocument, Database.In.CreateAndSaveDocument),
        (DocumentCache.Out.DeleteDocument, Database.In.DeleteDocument),
        (Database.Out.RequestedDocument, DocumentCache.In.RequestedDocument),
        (Database.Out.DocumentChanged, DocumentCache.In.DocumentChanged),
        (Database.Out.NewDocument, DocumentCache.In.DocumentCreated),
        (Database.Out.DocumentSaved, DocumentCache.In.DocumentSaved),
        (Database.Out.DocumentCreated, DocumentCache.In.DocumentSaved)
    ]

    def __init__(self, room): Peer.__init__(self, room)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *
from viewer import Viewer

class DocumentPathEnabler(Peer):

    def __init__(self, room, cb, enablePath, enablePathValue):
        Peer.__init__(self, room)
        self._cb = cb
        self._viewer = Viewer(room, self._enable, enablePath, raw=True)
        self._enablePathValue = enablePathValue
        self._countValue = False
        self._count = None
        self._operator = None

        if self._enablePathValue and self._enablePathValue.startswith('#'):
            self._countValue = True
            self._operator = self._enablePathValue[1:2]
            self._count = int(self._enablePathValue[2:])

    def _enable(self, value):
        if self._countValue:
            if value != None and isinstance(value, list):
                l = len(value)
                if (self._operator == '=' and l == self._count) or \
                   (self._operator == '>' and l > self._count) or \
                   (self._operator == '<' and l < self._count):
                    self._cb(True)
                    return

            self._cb(False)
            return

        if value == self._enablePathValue or unicode(value) == unicode(self._enablePathValue):
            self._cb(True)
            return

        self._cb(False)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *
from editDocument import EditDocument

class EditParty(Peer):

    def __init__(self, room, database=None, docId=None, dbClass=None, *args, **ka):
        Peer.__init__(self, room)

        from documentDatabase import DocumentDatabase
        from documentCache import DocumentCache
        from enableDocument import EnableDocument
        from debugger import Debugger
        from dialog import Dialog

        self._docId = docId

        self.add(DocumentDatabase(room))

        # Peers...
        self.add(Debugger(room))
        self.add(DocumentCache(room))
        self.add(EditDocument(room))
        self.add(EnableDocument(room))

        if dbClass is None:
            # try CouchDB as default
            from couchDB import CouchDB
            self.add(CouchDB(room, query=False, databaseName=database, **ka))
        else:
            self.add(dbClass(room, query=False, databaseName=database, **ka))

    def initialize(self):
        if self._docId != None:
            self._throw(EditDocument.In.Load, self._docId)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class PingPong(Peer):
    from ping import Ping
    from pong import Pong

    Routings = [
        (Ping.Out.Ping, Pong.In.Pong),
        (Pong.Out.Pong, Ping.In.Ping)
    ]

    def __init__(self, room): Peer.__init__(self, room)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class Database(Peer):
    from wallaby.pf.pillow import Pillow

    #Receiving pillows
    RequestDocument       = Pillow.In
    SaveDocument          = Pillow.In
    CreateAndSaveDocument = Pillow.In
    DeleteDocument        = Pillow.In
    CreateDocument        = Pillow.In

    #Sending pillows
    NewDocument           = Pillow.Out
    RequestedDocument     = Pillow.Out 
    DocumentChanged       = Pillow.Out
    ConnectionRefused     = Pillow.Out 
    ConnectionEstablished = Pillow.Out 
    DocumentNotFound      = Pillow.Out 
    DocumentNotDeleted    = Pillow.Out 
    DocumentNotSaved      = Pillow.Out
    DocumentDeleted       = Pillow.Out
    DocumentCreated       = Pillow.Out
    DocumentSaved         = Pillow.Out

    def __init__(self, *args):
        self._catch(Database.In.RequestDocument, self._getDocument)
        self._catch(Database.In.SaveDocument, self._setDocument)
        self._catch(Database.In.DeleteDocument, self._deleteDocument)
        self._catch(Database.In.CreateDocument, self._createDocument)
        self._catch(Database.In.CreateAndSaveDocument, self._createAndSaveDocument)

    def _getDocument(self, *args):
        pass

    def _setDocument(self, *args):
        pass

    def _deleteDocument(self, *args):
        pass

    def _createDocument(self, *args):
        pass



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class Dialog(Peer):
    Show = Pillow.In
    Hide = Pillow.In

    def __init__(self, room, dialog):
        Peer.__init__(self, room)
        self._dialog = dialog
        self._catch(Dialog.In.Show, self._show)
        self._catch(Dialog.In.Hide, self._hide)

    def _show(self, *args):
        self._dialog.show()
        self._dialog.raise_()

    def _hide(self, *args):
        self._dialog.hide()



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

class AbstractQuery(object):
    from wallaby.pf.pillow import Pillow

    __metaclass__ = Pillow
    #Receiving pillows
    Query = Pillow.In

    #Sending pillows
    Result = Pillow.OutState

    def __init__(self, *args):
        self._catch(AbstractQuery.In.Query, self._query)

    def _query(self, pillow, query):
        pass



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from editor import Editor

class ImageEditor(Editor):
    def __init__(self, *args, **ka):
        Editor.__init__(self, *args)

    def _fieldChanged(self):
        name = value = None
        if self._valueCallback:
            value, name = self._valueCallback()

        if (self._document):
            self._document.set(self._path, name)

            if name:
                self._document.setAttachment(name, value)

        self._throwFieldChanged(self._path)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from abstractQuery import AbstractQuery
from documentCache import DocumentCache
from editDocument import EditDocument
from multiViewer import MultiViewer
from tab import Tab

class MultiViewQuery(Peer):
    Description = ("Routing between a multi-view element (like a table) and a query-peer (like a database)",
                   '{ "dstRoom": null, "pillow": "EditDocument.In.Load", "unload": true }')

    Sending = [
        EditDocument.In.Load, 
        Tab.In.Select,
    ]

    Receiving = [
        MultiViewer.Out.Select
    ]

    Routings = [
        (AbstractQuery.Out.Result, MultiViewer.In.Data),
        (DocumentCache.Out.DocumentChanged, MultiViewer.In.DataChanged),
        (MultiViewer.Out.Query, AbstractQuery.In.Query)
    ]

    def __init__(self, room, dstRoom=None, pillow=EditDocument.In.Load, unload=True):
        Peer.__init__(self, room)
        self._pillow = pillow

        if dstRoom is None: dstRoom = room
        self._dstRoom = dstRoom
        self._unload = unload

        self._catch(MultiViewer.Out.Select, self._select)

    def rooms(self, pillow, feathers):
        rooms = Peer.rooms(self)
        if pillow == MultiViewer.Out.Select:
            rooms.add(self._dstRoom)

        return rooms

    def _select(self, pillow, feathers):
        id, tab, idx  = feathers

        if not self._unload and id == None:
            return

        if self._pillow != None:
            if isinstance(self._pillow, (list, tuple)):
                for p in self._pillow:
                    self._throw(self._dstRoom + ":" + p, id)
            else:
                self._throw(self._dstRoom + ":" + self._pillow, id)
        if tab != None:
            self._throw(self._dstRoom + ":" + Tab.In.Select, tab)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class IdQueryResultDocument(Peer):
    from documentIDQueryResult import DocumentIDQueryResult
    from documentCache import DocumentCache
    from database import Database

    Description = "Routing between a ID-based view widget and a document provider (like a document-cache)"

    Routings = [
        (DocumentCache.Out.RequestedDocument, DocumentIDQueryResult.In.RequestedDocument),
        (DocumentIDQueryResult.Out.RequestDocument, DocumentCache.In.RequestDocument)
    ]

    def __init__(self, room):
        Peer.__init__(self, room)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class EnableDocument(Peer):
    from editDocument import EditDocument
    from enabler import Enabler

    Routings = [
        (EditDocument.Out.State, Enabler.In.Enable)
    ]

    def __init__(self, room):
        Peer.__init__(self, room)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from uiPeer import *

class ListPeer(UIPeer):
    RequestedDocument = Pillow.In
    RequestDocument = Pillow.Out

    def __init__(self, room, cb, path, source):
        UIPeer.__init__(self, room, path)
        self._cb = cb
        self._source = source
        self._catch(ListPeer.In.RequestedDocument, self._requestedDocument)
        self._initialized = False

    def initialize(self):
        if self._initialized: return
        self._initialized = True

        self._throw(ListPeer.Out.RequestDocument, self._source)

    def _requestedDocument(self, pillow, document):
        if document.documentID == self._source:
            self._cb(document)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from multiViewer import MultiViewer

class IncrementalMultiViewer(MultiViewer):
    def __init__(self, *args, **ka):
        MultiViewer.__init__(self, *args, **ka)

        self._dct = {}
        self._IDs = []

    def _insertItems(self, f, elems):
        if self._delegate:
            self._delegate.beginInsertCB(f, len(elems))

        i = f
        for _id in elems:
            self._IDs.insert(i, _id)
            if _id in self._dct: self._dct[_id] += 1
            else: self._dct[_id] = 1
            i = i + 1

        if self._delegate:
            self._delegate.endInsertCB(f, len(elems))

    def _deleteItems(self, f, t):
        if self._delegate:
            self._delegate.beginDeleteCB(f, t)

        for i in range(f, t):
            if self._IDs[f] in self._dct:
                self._dct[self._IDs[f]] -= 1
                if self._dct[self._IDs[f]] == 0:
                    del self._dct[self._IDs[f]]
            else:
                print 'Error: Maybe duplicated entry in view'
            self._IDs.pop(f)

        if self._delegate:
            self._delegate.endDeleteCB(f, t)

    def _updateData(self, pillow, document):
        if document.documentID in self._IDs and self._delegate:
            self._delegate.updateCB(self._IDs.index(document.documentID))

   
    def _doRefreshData(self, queryResult):
        if queryResult:
            oldIDs = self._IDs

            self._queryResult = queryResult
            newIDs = queryResult.getDocumentIDs(range(queryResult.length()))

            c1 = len(oldIDs)
            c2 = len(newIDs)
            i1 = i2 = 0

            # print "View changed from", c1, "to", c2

            self._count = c2

            while i1 < c1 or i2 < c2:
                # not at end of any list
                if i1 < c1 and i2 < c2:
                    # item inserted
                    if oldIDs[i1] != newIDs[i2]:
                        if newIDs[i2] in self._dct:
                            self._deleteItems(i1, i1+1)
                            c1 = c1 - 1
                        else:
                            self._insertItems(i1, [newIDs[i2]])
                            i2 = i2 + 1
                            i1 = i1 + 1
                            c1 = c1 + 1
                    else:
                        if self._delegate:
                            self._delegate.updateCB(i1)
                        i2 = i2 + 1
                        i1 = i1 + 1

                # add item
                elif i2 < c2:
                    self._insertItems(i1, newIDs[i2:c2])
                    i2 = i2 + 1
                    i1 = i1 + 1
                    break

                # remove item
                elif i1 < c1:
                    self._deleteItems(i1, c1)
                    break

        if self._delegate:
            self._delegate.resetSelectionCB()



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class DocumentViewer(Peer):
    from documentCache import DocumentCache
    from viewer import Viewer

    Routings = [
        (DocumentCache.Out.RequestedDocument, Viewer.In.Document)
    ]

    def __init__(self, room):
        Peer.__init__(self, room)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from sm import *
import wallaby.FX as FX

from embeddedViewer import EmbeddedViewer
from documentChanger import DocumentChanger
from documentCache import DocumentCache
from multiViewer import MultiViewer
from database import Database
from viewer import Viewer
from editor import Editor

class EditDocument(SMPeer):
    Description = ("Provides a document edit/view logic", '{"alwaysEdit": false}')

    Load = Pillow.In
    LoadAndEdit = Pillow.In
    Edit = Pillow.In
    New = Pillow.In
    NewAndSave = Pillow.In
    Save = Pillow.In
    Remove = Pillow.In
    Rollback = Pillow.In
    Resolve = Pillow.In

    State = Pillow.OutState

    Dependencies = [
        "DocumentCache",
        "DocumentChanger"
    ]

    Sending = [
        Viewer.In.Document, 
        Viewer.In.Refresh, 
        MultiViewer.In.Refresh, 
        Editor.In.Conflict,
        Editor.In.Enable,
        DocumentCache.In.CreateShadowCopy,
        Database.In.CreateDocument,
        DocumentCache.In.RequestDocument,
        DocumentCache.In.StoreShadowCopy,
        DocumentCache.In.DeleteShadowCopy
    ]

    Receiving = [
        Database.Out.DocumentNotFound,
        Database.Out.DocumentDeleted,
        Database.Out.DocumentCreated,
        Editor.Out.FieldChanged,
        DocumentCache.Out.RequestedDocument,
        DocumentCache.Out.DoesNotExist,
        DocumentCache.Out.DocumentChanged,
        DocumentCache.In.DocumentCreated,
        DocumentCache.In.DocumentSaved
    ] 

    def __init__(self, room, **args):
        store = DocumentStore()
        states = []

        states.append(Undef(store, **args))
        states.append(Loading(store, **args))
        states.append(LoadingWillEdit(store, **args))
        states.append(Saving(store, **args))
        states.append(SavingNew(store, **args))
        states.append(View(store, **args))
        states.append(Edit(store, **args))
        states.append(Dirty(store, **args))
        states.append(Creating(store, **args))
        states.append(CreatingAndSave(store, **args))
        states.append(New(store, **args))

        SMPeer.__init__(self, room, EditDocument.Out.State, states, "Undef")

class DocumentStore:
    def __init__(self):
        self._documentID = None
        self._document = None

    @FX.Property
    def document():
        doc = "The document Property."
        def fget(self):
            return self._document
        def fset(self, value):
            self._document = value


    @FX.Property
    def documentID():
        doc = "The documentID Property."
        def fget(self):
            return self._documentID
        def fset(self, value):
            self._documentID = value

class Loadable:
    def __init__(self, **args):
        self._catch(EditDocument.In.Load, self._load)
        self._catch(EditDocument.In.LoadAndEdit, self._loadAndEdit)

        if not args.get("ignoreNotFound", False):
            self._setTransition(Database.Out.DocumentNotFound, "Undef")

        self._addRouting(Database.Out.DocumentCreated, MultiViewer.In.Refresh)
        self._addRouting(Database.Out.DocumentDeleted, MultiViewer.In.Refresh)

        self._config = args

    def _load(self, pillow, documentID):
        if documentID == None:
            self._switchState("Undef")
            return

        if 'alwaysEdit' in self._config and self._config['alwaysEdit']:
            self._switchState("LoadingWillEdit")
        else:
            self._switchState("Loading")

        self._store.documentID = documentID
        self._throw(DocumentCache.In.RequestDocument, self._store.documentID)

    def _loadAndEdit(self, pillow, documentID):
        if documentID == None:
            self._switchState("Undef")
            return

        if documentID == self._store.documentID and isinstance(self, Dirty):
            # already editing...
            # FX.debug("Skipping, because already in edit mode")
            return

        self._switchState("LoadingWillEdit")
        self._store.documentID = documentID
        self._throw(DocumentCache.In.RequestDocument, self._store.documentID)

    def _stateSwitched(self):
        self._throw(Viewer.In.Document, None)
        self._throw(Editor.In.Enable, False)

    def _create(self, pillow, feathers):
        self._switchState("Creating")

        from twisted.internet import reactor
        reactor.callLater(0, self.__create, pillow, feathers)

    def _createAndSave(self, pillow, feathers):
        self._switchState("CreatingAndSave")

        from twisted.internet import reactor
        reactor.callLater(0, self.__create, pillow, feathers)

    @defer.inlineCallbacks
    def __create(self, pillow, feathers):
        if isinstance(feathers, str) or isinstance(feathers, unicode):
            try:
                feathers = yield DocumentChanger.transform(None, feathers)
            except Exception as e:
                feathers = None
                print "ERROR - Exception: " + str(e)

        if type(feathers) is dict:
            self._throw(Database.In.CreateDocument, feathers)
        else:
            self._throw(Database.In.CreateDocument, None)

class OnShadowCopy:
    def __init__(self, **args):
        self._catch(EditDocument.In.Save, self._save)
        self._catch(DocumentCache.Out.DocumentChanged, self._docChanged)

    def _docChanged(self, pillow, doc):
        if doc.documentID != self._store.documentID:
            return

        if isinstance(self, (Dirty, New)) and not isinstance(self, Edit):
            self._throw(Editor.In.Conflict, doc)
        else:
            if 'alwaysEdit' in self._config and self._config['alwaysEdit']:
                self._switchState("LoadingWillEdit")
            else:
                self._switchState("Loading")

            self._store.documentID = doc.documentID
            self._throw(DocumentCache.In.RequestDocument, doc.documentID)

    def _stateSwitched(self):
        self._throw(Editor.In.Enable, True)

    def _save(self, *args):
        self._switchState("Saving")
        self._throw(DocumentCache.In.StoreShadowCopy, self._store.documentID)

class Undef(Loadable, State):
    def __init__(self, store, **args):
        State.__init__(self)
        Loadable.__init__(self, **args)
        self._store = store
        self._catch(EditDocument.In.New, self._create)
        self._catch(EditDocument.In.NewAndSave, self._createAndSave)

    def _stateSwitched(self):
        Loadable._stateSwitched(self)

class Loading(Loadable, State):
    def __init__(self, store, **args):
        State.__init__(self)
        Loadable.__init__(self, **args)
        self._store = store
        self._catch(DocumentCache.Out.RequestedDocument, self._document)
        if 'alwaysEdit' in args and args['alwaysEdit']:
            self._setTransition(DocumentCache.Out.RequestedDocument, "Edit")
        else:
            self._setTransition(DocumentCache.Out.RequestedDocument, "View")

    def _stateSwitched(self):
        self._throw(Editor.In.Enable, False)

    def _document(self, pillow, document):
        if self._store.documentID == document.documentID:
            self._store.document = document
            self._throw(Viewer.In.Document, document)

class LoadingWillEdit(Loadable, State):
    def __init__(self, store, **args):
        State.__init__(self)
        Loadable.__init__(self, **args)
        self._store = store
        self._catch(DocumentCache.Out.RequestedDocument, self._document)
        self._setTransition(DocumentCache.Out.RequestedDocument, "Edit")

    def _stateSwitched(self):
        self._throw(Editor.In.Enable, False)

    def _document(self, pillow, document):
        if self._store.documentID == document.documentID:
            self._throw(DocumentCache.In.ReplaceShadowCopy, self._store.documentID)
            self._store.document = document
            self._throw(Viewer.In.Document, document)

class Saving(OnShadowCopy, State):
    def __init__(self, store, **args):
        State.__init__(self)
        OnShadowCopy.__init__(self, **args)
        self._store = store
        self._config = args
        self._catch(DocumentCache.Out.DocumentSaved, self._documentSaved)

    def _stateSwitched(self):
        self._throw(Editor.In.Enable, False)

    def _documentSaved(self, pillow, documentID):
        if self._store.documentID == documentID:
            if 'alwaysEdit' in self._config and self._config['alwaysEdit']:
                self._throw(DocumentCache.In.CreateShadowCopy, self._store.documentID)
                self._switchState("Edit")
            else:
                self._switchState("View")

class SavingNew(Saving):
    def _documentSaved(self, pillow, documentID):
        Saving._documentSaved(self, pillow, documentID)
        if self._store.documentID == documentID:
            self._throw(DocumentCache.In.RequestDocument, self._store.documentID)
            self._throw(MultiViewer.In.Refresh, self._store.documentID)

class View(Loading):
    def __init__(self, store, **args):
        Loading.__init__(self, store, **args)
        self._catch(DocumentCache.Out.DocumentChanged, self._document)
        self._catch(EditDocument.In.Edit, self._edit)
        self._catch(EditDocument.In.New, self._create)
        self._catch(EditDocument.In.NewAndSave, self._createAndSave)
        self._catch(EditDocument.In.Remove, self._remove)
        self._addRouting(DocumentChanger.Out.SelectionChanged, Viewer.In.Refresh)

    def _stateSwitched(self):
        self._throw(Editor.In.Enable, False)

    def _edit(self, *args):
        self._switchState("Edit")
        self._throw(DocumentCache.In.CreateShadowCopy, self._store.documentID)

    def _remove(self, *args):
        self._switchState("Undef")
        self._throw(DocumentCache.In.DeleteDocument, self._store.documentID)

class Dirty(OnShadowCopy, State):
    def __init__(self, store, **args):
        State.__init__(self)
        OnShadowCopy.__init__(self, **args)
        self._store = store
        self._config = args

        self._catch(EditDocument.In.Rollback, self._rollback)
        self._catch(EditDocument.In.Resolve, self._resolve)
        self._addRouting(Editor.Out.FieldChanged, Viewer.In.Refresh)

        self._addRouting(DocumentChanger.Out.SelectionChanged, Viewer.In.Refresh)

    def _resolve(self, *args):
        if self._store.document:
            self._store.document.resolve()
            self._save(*args)

    def _rollback(self, *args):
        self._switchState("View")
        self._throw(DocumentCache.In.DeleteShadowCopy, self._store.documentID)

        if 'alwaysEdit' in self._config and self._config['alwaysEdit']:
            self._throw(EditDocument.In.LoadAndEdit, self._store.documentID)
        else:
            self._throw(EditDocument.In.Load, self._store.documentID)

class Edit(Dirty, Loadable):
    def __init__(self, store, **args):
        Dirty.__init__(self, store, **args)
        Loadable.__init__(self, **args)
        self._addRouting(Editor.Out.FieldChanged, Viewer.In.Refresh)
        self._setTransition(Editor.Out.FieldChanged, "Dirty")
        self._setTransition(DocumentCache.Out.DoesNotExist, "Undef")

        if 'alwaysEdit' in args and args['alwaysEdit']:
            self._catch(EditDocument.In.New, self._create)
            self._catch(EditDocument.In.NewAndSave, self._createAndSave)
            self._catch(EditDocument.In.Remove, self._remove)

    def _remove(self, *args):
        self._switchState("Undef")
        self._throw(DocumentCache.In.DeleteDocument, self._store.documentID)

class Creating(State):
    def __init__(self, store, **args):
        State.__init__(self)
        self._store = store
        self._catch(DocumentCache.Out.DocumentCreated, self._new)
        self._addRouting(Editor.Out.FieldChanged, Viewer.In.Refresh)

    def _new(self, pillow, document):
        self._switchState("New")
        self._store.documentID = document.documentID
        self._throw(Viewer.In.Document, document)

class CreatingAndSave(State):
    def __init__(self, store, **args):
        State.__init__(self)
        self._store = store
        self._catch(DocumentCache.Out.DocumentCreated, self._new)

    def _new(self, pillow, document):
        self._switchState("SavingNew")
        self._store.documentID = document.documentID
        self._throw(DocumentCache.In.StoreShadowCopy, self._store.documentID)

class New(OnShadowCopy, State):
    def __init__(self, store, **args):
        State.__init__(self)
        OnShadowCopy.__init__(self, **args)
        self._store = store
        self._catch(EditDocument.In.Rollback, self._rollback)
        self._addRouting(Editor.Out.FieldChanged, Viewer.In.Refresh)

    def _save(self, *args):
        self._switchState("SavingNew")
        self._throw(DocumentCache.In.StoreShadowCopy, self._store.documentID)

    def _rollback(self, *args):
        self._switchState("Undef")
        self._throw(DocumentCache.In.DeleteShadowCopy, self._store.documentID)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from viewer import Viewer

class ImageViewer(Viewer):
    def __init__(self, room, cb, path):
        Viewer.__init__(self, room, cb, path)

    def _refresh(self, pillow, path):
        from wallaby.common.pathHelper import SelectionNotSpecified
        from twisted.internet import defer

        if self._path and Viewer.matchPath(self._path, path, mode=Viewer.OUTER):
            d = None

            try:
                if self._document:
                    name = self._document.get(self._path)

                    if name == None: name = self._path

                    if name != None:
                        d = defer.maybeDeferred(self._document.deferredGetAttachment, name)

                if d and self._credentials:
                    credential = self._credentials.getFlat(self._path)
                    if credential == None or 'view' in credential:
                        self._cb(d, name)
                else:
                    d = defer.Deferred()
                    d.callback(None)
                    self._cb(d, None)

            except SelectionNotSpecified as e:
                d = defer.Deferred()
                d.callback(None)
                self._cb(d, None)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from wallaby.pf.peer.database import Database
from wallaby.pf.peer.viewer import Viewer
from wallaby.pf.peer.editor import Editor
from wallaby.pf.pillow import Pillow
from wallaby.common.pathHelper import *
from twisted.internet import defer
from datetime import date, datetime
import copy, re

class DocumentChanger(Peer):
    Receiving = [
        Viewer.In.Document
    ]

    Sending = [
        Database.In.SaveDocument
    ]

    Change = Pillow.In
    Edit = Pillow.In
    Clone = Pillow.In
    SubClone = Pillow.In
    Embed = Pillow.In

    InsertRow = Pillow.In
    RemoveRow = Pillow.In

    RowInserted = Pillow.Out

    InsertKey = Pillow.In
    RemoveKey = Pillow.In

    MoveUp = Pillow.In
    MoveDown = Pillow.In

    Select = Pillow.In
    SelectionChanged = Pillow.Out

    def __init__(self, room, dstRoom=None):
        Peer.__init__(self, room)
        self._document = None

        self._dstRoom = None

        if dstRoom != None: 
            from wallaby.pf.room import House
            self._dstRoom = House.get(dstRoom)

        self._catch(DocumentChanger.In.Change, self._change)
        self._catch(DocumentChanger.In.Edit, self._edit)
        self._catch(DocumentChanger.In.Clone, self._clone)
        self._catch(DocumentChanger.In.SubClone, self._subClone)
        self._catch(DocumentChanger.In.Embed, self._embed)
        self._catch(Viewer.In.Document, self._setDocument)

        self._catch(DocumentChanger.In.InsertRow, self._insertRow)
        self._catch(DocumentChanger.In.RemoveRow, self._removeRow)

        self._catch(DocumentChanger.In.InsertKey, self._insertKey)
        self._catch(DocumentChanger.In.RemoveKey, self._removeKey)

        self._catch(DocumentChanger.In.MoveUp, self._up)
        self._catch(DocumentChanger.In.MoveDown, self._down)

        self._catch(DocumentChanger.In.Select, self._select)

    #TODO: Add support for Enabler

    def _select(self, pillow, (path, idx) ):
        if not self._document: return

        path = path + ".*"
    
        try:
            currentIdx = PathHelper.getValue(self._document.selection, path + '._selection') 
            if currentIdx == idx: 
                # print path, ":", idx, "Already selected"
                return
        except:
            pass

        self._document.select(path, idx)
        path = path.replace('__root__.', '')
        self._throw(DocumentChanger.Out.SelectionChanged, path)

    def _up(self, pillow, path): 
        data = self._document.get(path, asList=True) 
        if data == None: return 

        sel = PathHelper.getValue(self._document.selection, path + '.*._selection') 
        if sel == None or sel == 0: return 
 
        data[sel-1], data[sel] = data[sel], data[sel-1]  
        self._throw(Editor.Out.FieldChanged, Editor.translatePath(self._document, path))

    def _down(self, pillow, path):
        data = self._document.get(path, asList=True)
        if data == None: return

        sel = PathHelper.getValue(self._document.selection, path + '.*._selection')
        if sel == None or sel == len(data)-1: return

        data[sel], data[sel+1] = data[sel+1], data[sel]
        self._throw(Editor.Out.FieldChanged, Editor.translatePath(self._document, path))
 
    def _embed(self, pillow, path):
        if not path or not self._document:
            return

        self._dstRoom.throw(DocumentChanger.In.InsertRow, (path, copy.deepcopy(self._document._data)) )

    def _subClone(self, pillow, doc):
        from twisted.internet import reactor
        reactor.callLater(0, self.__subClone, pillow, doc)

    def _insertKey(self, *args):
        self.__insert(*args, isList=False)

    def _removeKey(self, *args):
        self.__remove(*args, isList=False)

    def _insertRow(self, *args):
        self.__insert(*args, isList=True)

    def _removeRow(self, *args):
        self.__remove(*args, isList=True)

    def __insert(self, pillow, path, isList=True):
        if not path or not self._document:
            return

        doc = None
        key = '__new__'

        if isinstance(path, tuple) or isinstance(path, list):
            path, doc = path
        elif isinstance(path, dict):
            cfg = path
            path = cfg.get('path', None)
            key  = cfg.get('key', None)
            doc  = cfg.get('value', None)

        try:
            if isList:
                data = self._document.get(path, asList=True)
            else:
                data = self._document.get(path, asDict=True)
        except Exception as e:
            return

        if data == None:
            if isList:
                data = []
            else:
                data = {}

            try:
                self._document.set(path, data)
            except Exception as e:
                return

        if not isList and key in data:
            return

        try:
            sel = PathHelper.getValue(self._document.selection, path + '.*._selection')
        except SelectionNotSpecified:
            sel = None

        if sel == None:
            if isList:
                sel = len(data)
            else:
                sel = key

        if isList:
            data.insert(sel, doc)
        else:
            data[key] = doc


        self._throw(Editor.Out.FieldChanged, Editor.translatePath(self._document, path))

        if isList:
            self._select(None, (path, sel))

        if not isList:
            sel = key
            self._select(None, (path, sel))

        self._throw(DocumentChanger.Out.RowInserted, (Editor.translatePath(self._document, path), sel))

    def __remove(self, pillow, path, isList=True):
        if not path or not self._document:
            # TODO: debug out
            return

        if isList:
            data = self._document.get(path, asList=True)
        else:
            data = self._document.get(path, asDict=True)

        if data == None:
            # TODO: debug out
            return

        sel = PathHelper.getValue(self._document.selection, path + '.*._selection')
        if sel != None:
            if not isList:
                if not sel in data: return

                keys = sorted(data.keys())
                idx = keys.index(sel)

                if idx == len(keys)-1:
                    idx -= 1
            else:
                if sel >= len(data): return

            del data[sel]

            self._throw(Editor.Out.FieldChanged, Editor.translatePath(self._document, path))

            if isList and sel == len(data) and sel != 0:
                self._select(None, (path, sel-1))

            if not isList:
                keys = sorted(data.keys())
                if idx >= 0 and idx < len(keys):
                    self._select(None, (path, keys[idx]))


    @defer.inlineCallbacks
    def __subClone(self, pillow, feathers):
        if not feathers or not self._document:
            return

        path = feathers
        doc = None

        if ';' in feathers:
            path, doc = feathers.split(';')

        item = self._document.get(path)
        if not item: return

        from wallaby.common.document import Document
        newDoc = Document(data=copy.deepcopy(item))
        newDoc.resetDocumentID()

        if doc:
            yield DocumentChanger.transform(newDoc, doc)

        self._throw(Database.In.SaveDocument, (newDoc, None) )

    transformations = {
        "Y": lambda _: datetime.now().year,
        "M": lambda _: datetime.now().month,
        "D": lambda _: datetime.now().day,
        "h": lambda _: datetime.now().hour,
        "m": lambda _: datetime.now().minute,
        "s": lambda _: datetime.now().second
    } 

    @staticmethod
    def registerToken(token, callback):
        if token in DocumentChanger.transformations:
            print "ERROR - token", token, "already registered"
            return

        DocumentChanger.transformations[token] = callback

    @staticmethod
    def reregisterToken(token, callback):
        DocumentChanger.transformations[token] = callback

    @staticmethod
    def transform(newDoc, doc):
        if not doc:
            return False

        d = defer.Deferred()
        from twisted.internet import reactor
        reactor.callLater(0, DocumentChanger.__doTransform, d, newDoc, doc)
        return d

    @staticmethod
    @defer.inlineCallbacks
    def __doTransform(d, newDoc, doc):
        import json
        if isinstance(doc, (unicode, str)):
            while True:
                m = re.search(r"%(\w+)", doc)
                if m == None: break

                token = m.group(1)
                cb = DocumentChanger.transformations.get(m.group(1), lambda _: "null")

                from twisted.internet import reactor, task
                replacement = yield task.deferLater(reactor, 0, cb, newDoc)

                if not isinstance(replacement, str) and not isinstance(replacement, unicode):
                    replacement = json.dumps(replacement)

                doc = doc.replace('%' + token, unicode(replacement), 1)

            try:
                doc = json.loads(doc)
            except Exception as e:
                print "DocumentChanger", e
                d.errback("JSON error " + unicode(e) + " in: "+ doc)
                return

        if newDoc == None:
            newDoc = doc
        else:
            for key, val in doc.items():
                newDoc.set(key, val)

        d.callback(newDoc)

    def _clone(self, pillow, doc):
        from twisted.internet import reactor
        reactor.callLater(0, self.__clone, pillow, doc)

    @defer.inlineCallbacks
    def __clone(self, pillow, doc):
        if not doc or not self._document:
            return

        referenceID = None

        if ';' in doc:
            referenceID, doc = doc.split(';')

        newDoc = self._document.clone()
        newDoc.resetDocumentID()

        yield DocumentChanger.transform(newDoc, doc)

        if referenceID != None:
            newDoc.set(referenceID, self._document.documentID)

        self._throw(Database.In.SaveDocument, (newDoc, self._document) )

    def _change(self, pillow, doc):
        from twisted.internet import reactor
        reactor.callLater(0, self.__change, pillow, doc)

    def _edit(self, pillow, doc):
        from twisted.internet import reactor
        reactor.callLater(0, self.__edit, pillow, doc)

    def __edit(self, pillow, doc):
        if not doc or not self._document:
            return

        if not isinstance(doc, dict):
            try:
                import json
                doc = json.loads(doc)
            except Exception as e:
                print "JSON error " + unicode(e) + " in: "+ doc
                return

        for k, v in doc.items():
            self._document.set(k, v)
            self._throw(Editor.Out.FieldChanged, Editor.translatePath(self._document, k))

    @defer.inlineCallbacks
    def __change(self, pillow, doc):
        if not doc or not self._document:
            return

        newDoc = copy.deepcopy(self._document)
        yield DocumentChanger.transform(newDoc, doc)

        self._throw(Database.In.SaveDocument, (newDoc, self._document) )

    def _setDstDocument(self, pillow, document):
        self._dstDocument = document

    def _setDocument(self, pillow, document):
        self._document = document



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class ViewDocumentFromDatabase(Peer):
    from documentChanger import DocumentChanger
    from database import Database
    from viewer import Viewer

    Load = Pillow.In 

    Routings = [
        (Database.Out.RequestedDocument, Viewer.In.Document),
        (DocumentChanger.Out.SelectionChanged, Viewer.In.Refresh)
    ]

    def __init__(self, room):
        Peer.__init__(self, room)
        self._addRouting(ViewDocumentFromDatabase.In.Load, Database.In.RequestDocument)

    def initialize(self):
        pass 






# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from editor import Editor

class DoubleEditor(Editor):
    def _fieldChanged(self):
        from wallaby.common.pathHelper import SelectionNotSpecified

        value = 0
        if self._valueCallback:
            value = self._valueCallback()
        if (self._document):
            try:
                self._document.set(self._path, value) 
                self._throwFieldChanged(self._path)
            except SelectionNotSpecified as e:
                self._setReadOnly(True)




# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class SdEdit(Peer):
    def __init__(self, room, host='localhost', port=60001, actors = None):
        Peer.__init__(self, room)

        self._roomName = room

        self._catch('*', self._message, passCaller=True)

        if actors == None: actors = {}

        self._host = host
        self._port = port

        self._actors = actors

        from wallaby.pf.log.sdeditChannelFactory import SDEditChannelFactory
        self._factory = SDEditChannelFactory()

        self._initialized = False

    def initialize(self):
        if self._initialized: return
        self._initialized = True

        pillows = House.rooms[self._roomName].registeredPillows

        self._peers = {}
        self._pillows = {}
        self._translation = {}
        for key, callbacks in pillows.items():
            for (callback, thrower) in callbacks:
                className = callback.im_self.__class__.__name__
                if len(self._actors) == 0 or className in self._actors:
                    if className in self._actors:
                        trans = self._actors[className]

                        if isinstance(trans, str) or isinstance(trans, unicode):
                            self._translation[className] = trans
                            className = trans
                        else:
                            if not trans:
                                pass 

                    if not className in self._peers: self._peers[className] = {}
                    if not key in self._pillows: self._pillows[key] = {}

                    self._peers[className][key] = True
                    self._pillows[key][className] = True

        for peer in self._actors:
            if peer in self._translation:
                peer = self._translation[peer]

            if peer in self._actors:
                trans = self._actors[peer]

                if isinstance(trans, str) or isinstance(trans, unicode):
                    self._translation[peer] = trans
                    peer = trans

            if peer not in self._peers:
                self._peers[peer] = {}


        self._factory.configure(peers=self._peers, room=self._roomName, pillows=self._pillows, translations=self._translation)

        from twisted.internet import reactor
        reactor.connectTCP(self._host, self._port, self._factory)

    def _message(self, pillow, feathers, caller=None):
        self._factory.sendMessage(caller.__class__.__name__, pillows, feathers)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from setuptools import setup, find_packages

setup(name='wallaby-base',
      version='0.1.43',
      url='https://github.com/FreshXOpenSource/wallaby-base',
      author='FreshX GbR',
      author_email='wallaby@freshx.de',
      license='BSD',
      description='Core package for wallaby.',
      long_description=open('README.md').read(),
      package_data={'': ['LICENSE', 'AUTHORS', 'README.md']},
      classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Twisted',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Application Frameworks'
      ],
      packages=find_packages('.'),
      # dependencies to couchdb and elasticsearch are for test-cases only. 
      install_requires=['wallaby-plugin-couchdb', 'wallaby-plugin-elasticsearch', 'setuptools-git']
  )



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class Credentials(Peer):  
    # Request credentials for dynamic peers
    Update     = Pillow.In

    # The requested (or unrequested) credential
    Credential = Pillow.Out
     
    def __init__(self, room):
        Peer.__init__(self, room)
        self._catch(Credentials.In.Update, self._update)

    def _update(self, *args):
        pass




# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class CredentialsConsumer(Peer):
    Credential = Pillow.InState

    def __init__(self, callback=None):
        room = "__CREDENTIALS__"

        Peer.__init__(self, room)
        self._roomName = room
        self._callback = callback
        self._credentials = None

        # Register for credential documents
        self._catch(CredentialsConsumer.In.Credential, self._newCredentials)

    def _newCredentials(self, pillow, feathers):
        if self._callback:
            self._callback(pillow, feathers)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class DocumentCredentials(Peer):
    # Message to request the credential document from backend
    RequestDocument = Pillow.Out
    SaveDocument = Pillow.Out

    # The requested credential document
    RequestedDocument = Pillow.In
    DocumentNotFound = Pillow.In
    DocumentCreated = Pillow.In

    Singleton = None

    def __init__(self, room, documentID="credentials"):
        Peer.__init__(self, room)
        self._documentID = documentID
        self._catch(DocumentCredentials.In.RequestedDocument, self._requestedDocument)
        self._catch(DocumentCredentials.In.DocumentNotFound, self._docNotFound)
        self._catch(DocumentCredentials.In.DocumentCreated, self._docCreated)

        DocumentCredentials.Singleton = self

        self._initialized = False

    @staticmethod
    def request():
        if DocumentCredentials.Singleton == None: return
        DocumentCredentials.Singleton.reinitialize()

    def _docCreated(self, pillow, id):
        if id != self._documentID: return
        self._update()

        # self._throw(DocumentCredentials.Out.RequestDocument, self._documentID)

    def _docNotFound(self, pillow, id):
        if id != self._documentID: return
        from wallaby.common.document import Document
        self._credentials = Document(self._documentID)
        self._throw(DocumentCredentials.Out.SaveDocument, (self._credentials, None))

        self._requestedDocument(None, self._credentials)

    def reinitialize(self):
        self._initialized = False
        self.initialize()

    def initialize(self):
        if self._initialized: return
        self._initialized = True

        self._throw(DocumentCredentials.Out.RequestDocument, self._documentID)

    def _requestedDocument(self, pillow, document):
        if document.documentID == self._documentID:
            self._credentials = document
            self._update()

    def _update(self, *args):
        from credentials import Credentials
        self._throw(Credentials.Out.Credential, self._credentials)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

import wallaby.backends.couchdb as couch
from wallaby.pf.peer.credentialsParty import *
from wallaby.pf.peer.configParty import *
from wallaby.pf.peer.credentials import *
from wallaby.pf.room import *
import wallaby.FX as FX

from wallaby.pf.peer.viewer import *
from wallaby.pf.peer.editor import *

class BaseApp(object):
    def __init__(self, quitCB, dbName='default', embedded=False, appName=None, *args, **ka):
        self._dbName = dbName

        self._debuggedRooms = []
        self._debuggers = []
        self._quitCB = quitCB

        self._credentialsInitialized = False

        if not embedded: 
            CredentialsParty(dbName)

            if appName == "inspector":
                ConfigParty("__CONFIG__", dbName, docId="WallabyApp2", inspectorDb=None)
            else:
                ConfigParty("__CONFIG__", dbName, docId="WallabyApp2", inspectorDb="bootstrap")

            House.catch("__CONFIG__", Viewer.In.Refresh, self._configRefreshed)
            House.catch("__CONFIG__", Viewer.In.Document, self._configRefreshed)
            House.catch("__CREDENTIALS__", Credentials.Out.Credential, self._credentialsArrived)

            House.catch("__CONFIG__", Viewer.In.RefreshDone, self._initializeCredentials)

        # House.get("__CONFIG__").initialize()
        # House.get("__CONFIG__").initializePeers()

    def _configRefreshed(self, pillow, feathers):
        if feathers != None:
            House.throw("__CONFIG__:" + Viewer.In.RefreshDone, None)

    def _initializeCredentials(self, pillow, feathers):
        if not self._credentialsInitialized:
            self._credentialsInitialized = True
            House.get("__CREDENTIALS__").initialize()
            House.get("__CREDENTIALS__").initializePeers()

    def quit(self):
        self._quitCB()

    def _credentialsArrived(self, pillow, feathers):
        pass

    def authenticated(self, username=None, password=None, options=None):
        d = defer.Deferred()
        from twisted.internet import reactor
        reactor.callLater(0, self._authenticated, username, password, options, d)
        return d
    
    @defer.inlineCallbacks
    def _authenticated(self, username, password, options, d):
        self.setConnectionSettings(options)

        if username != None:
            couch.Database.setLoginForDatabase(self.dbName(), username, password)

        authorized = False
        try:
            authorized = yield couch.Database.testConnectionToDatabase(self.dbName())
        except Exception as e:
            print "Exception while login", e 

        if authorized:
            options.username = username
            options.password = password
            self.setConnectionSettings(options) # reset connection settings

            from twisted.internet import reactor
            reactor.callLater(0, couch.Database.getDatabase(self.dbName()).changes)
            d.callback(True)
        else:
            d.callback(False)

    def run(self, checkRoom=None):
        dbname = self.dbName()
        if dbname == None:
            return

        for debugger in self._debuggers:
            debugger.destroy()

        # House.get("__CONFIG__").initialize()
        # House.get("__CONFIG__").initializePeers()
        House.get("__CONFIG__").configUpdated(None, force=True)

        for room in self._debuggedRooms:
            from wallaby.pf.peer.debugger import Debugger
            self._debuggers.append(Debugger(room, "*"))

        if checkRoom != None:
            House.get(checkRoom).check(suggest=True)

        self.isRunning()

    def isRunning(self):
        pass

    def setConnectionSettings(self, options):
        pass

    def dbName(self):
        return self._dbName



#!/usr/bin/env python
# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.


from peer import *

class CredentialsParty(Peer):
    def __init__(self, database=None, dbClass=None, *args, **ka):
        room = "__CREDENTIALS__"
        Peer.__init__(self, room)

        from documentDatabase import DocumentDatabase
        from documentCache import DocumentCache
        from documentCredentialsDocument import DocumentCredentialsDocument
        from documentCredentials import DocumentCredentials
        from credentialsRouting import CredentialsRouting
        from debugger import Debugger

        # Routings...
        self.add(DocumentDatabase(room))
        self.add(DocumentCredentialsDocument(room))
        self.add(CredentialsRouting(room))

        # Peers...
        self.add(Debugger(room))
        self.add(DocumentCache(room))
        self.add(DocumentCredentials(room))

        if dbClass is None:
            from couchDB import CouchDB
            self.add(CouchDB(room, query=False, databaseName=database, **ka))
        else:
            self.add(dbClass(room, query=False, databaseName=database, **ka))





# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class DocumentCredentialsDocument(Peer):
    from documentCache import DocumentCache
    from documentCredentials import DocumentCredentials
    from database import Database

    Routings = [
        (DocumentCache.Out.RequestedDocument, DocumentCredentials.In.RequestedDocument),
        (Database.Out.DocumentNotFound, DocumentCredentials.In.DocumentNotFound),
        (Database.Out.DocumentCreated, DocumentCredentials.In.DocumentCreated),
        (DocumentCredentials.Out.RequestDocument, DocumentCache.In.RequestDocument),
        (DocumentCredentials.Out.SaveDocument, Database.In.SaveDocument)
    ]

    def __init__(self, room):
        Peer.__init__(self, room)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class PayloadCallback(Peer):
    def __init__(self, room, pillow, cb, raw=False):
        Peer.__init__(self, room)
        self._cb = cb
        self._raw = raw
        self._catch(pillow, self._message)

    def _message(self, pillow, feather):
        if self._raw:
            self._cb(feather)
        else:
            self._cb(unicode(feather))



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from editor import Editor

class ListEditor(Editor):
    def __init__(self, *args, **ka):
        Editor.__init__(self, *args, **ka)

    def _fieldChanged(self):
        value = False
        if self._valueCallback:
            checkBox = self._valueCallback()
            value = checkBox.value
            if (self._document):
                listVal = self._document.get(self._path)
                if checkBox.isChecked():
                    if isinstance(listVal, list) or isinstance(listVal, tuple):
                        if value not in listVal:
                            listVal.append(value)
                    else:
                        listVal = [value]
                    if value in ('true', 'True'):
                        listVal = True
                    elif value in ('false', 'False'):
                        listVal = False 
                        
                else:
                    if isinstance(listVal, list) or isinstance(listVal, tuple): 
                        for i in range(len(listVal)):
                            if listVal[i] == value:
                                listVal.pop(i)
                                break
                    else:
                        listVal = False

                    if value in ('true', 'True'):
                        listVal = False
                    elif value in ('false', 'False'):
                        listVal = True 

                self._document.set(self._path, listVal) 

        self._throwFieldChanged(self._path)



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

from editDocument import EditDocument
from documentCache import DocumentCache
from documentChanger import DocumentChanger
from viewer import Viewer

class ViewDocument(Peer):
    Description = "View a document from the document cache"

    Routings = [
        (DocumentChanger.Out.SelectionChanged, Viewer.In.Refresh)
    ]

    Receiving = [
        EditDocument.In.Load,
        DocumentCache.Out.RequestedDocument,
        DocumentCache.Out.DocumentChanged
    ]

    Sending  = [
        Viewer.In.Document,
        EditDocument.Out.State,
        DocumentCache.In.RequestDocument
    ]

    Dependencies = [
        "DocumentCache",
        "DocumentChanger"
    ]

    def __init__(self, room):
        Peer.__init__(self, room)
        self._catch(EditDocument.In.Load, self._load)
        self._catch(DocumentCache.Out.RequestedDocument, self._document)
        self._catch(DocumentCache.Out.DocumentChanged, self._document)
        self._id = None

    def _load(self, pillow, docid):
        if docid == None:
            self._throw(Viewer.In.Document, None)
            self._throw(EditDocument.Out.State, 'Undef')
            return

        self._id = docid
        self._throw(DocumentCache.In.RequestDocument, docid)
        self._throw(EditDocument.Out.State, 'View')

    def _document(self, pillow, doc):
        from wallaby.common.document import Document
        if doc and isinstance(doc, Document) and doc.documentID == self._id:
            self._throw(Viewer.In.Document, doc)

    def initialize(self):
        self._throw(Viewer.In.Document, None)
        self._throw(EditDocument.Out.State, 'Undef')






# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from peer import *

class Ping(Peer):
    Ping = Pillow.InOut

    def __init__(self, *args):
        Peer.__init__(self, *args)
        self._catch(Ping.In.Ping, self.message)

    def message(self, pillow, feathers):
        self._throw(Ping.Out.Ping, int(feathers)+1);



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

from twisted.internet import defer, threads, task
from collections import deque

def Callable(fn):
    from functools import wraps
    @wraps(fn)
    def wrapper(self, *args, **kw):
        from twisted.internet import reactor
        return task.deferLater(reactor, 0, self.lockedCall, fn, self, *args, **kw)
    return wrapper

class Service(object):
    def __init__(self):
        # lock queue
        self.__callables = deque()
        self.__lock = defer.DeferredLock()

    def __unlock(self, val):
        # print "Release lock"
        self.__lock.release()
        return val

    # call a synchronous function in an asynchronous manner 
    def lockedCall(self, f, *args, **kw):
        # add function and arguments to lock queue
        self.__callables.append( (f, args, kw) )

        # acquire lock and call the passed function with arguments 
        # print "Wait for lock"
        d = self.__lock.acquire()
        d.addCallback(self.__call)
        return d

    # call the next function with arguments in the lock queue
    def __call(self, _): 
        # print "Acquire Lock"
        f, args, kw = self.__callables.popleft()
        d = threads.deferToThread(f, *args, **kw)
        
        # unlock definitely (even on error)
        d.addBoth(self.__unlock)
        return d



# Copyright (c) by it's authors. 
# Some rights reserved. See LICENSE, AUTHORS.

import wallaby.FX as FX
import sys, re

class SelectionNotSpecified(Exception):
    pass

class SelectionOutOfRange(SelectionNotSpecified): #TODO: Change inheritance
    pass

class PathHelper:
    @staticmethod
    def getSelection(path, selection=None):
        resPath = selectionPath = ""

        if path != None and re.search("__root__", path):
            path = path.replace("__root__.", "").replace("__root__", "")

        rest = str(path)

        while rest != None and len(rest) > 0:
            lst = rest.split('.', 1)
            if len(lst) == 2:
                part, rest = lst
            else:
                part, rest = lst[0], None

            selectionPath = selectionPath+part+'.'

            if part == '*':
                part = PathHelper.getValue(selection, selectionPath+'_selection')

                if part == None:
                    raise SelectionNotSpecified()

            resPath = resPath + unicode(part) + "."

        return re.sub("\.$", "", resPath)

    @staticmethod
    def getValue(base, path, selection=None, asList=False, asDict=False, wrap=False):
        if base != None:
            if path != None:
                grp = re.match(r"\"(.*)\"", path)
                if grp:
                    return grp.group(1)

            pos = base
            selectionPath = ""

            if path != None and re.search("__root__", path):
                path = path.replace("__root__.", "").replace("__root__", "")

            rest = str(path)

            while rest != None and len(rest) > 0:
                lst = rest.split('.', 1)
                if len(lst) == 2:
                    part, rest = lst
                else:
                    part, rest = lst[0], None

                selectionPath = selectionPath+part+'.'

                if isinstance(pos,dict):
                    if part in pos:
                        pos = pos[part]
                    elif selection != None and part == '*':
                        part = PathHelper.getValue(selection, selectionPath+'_selection')

                        if part == None:
                            raise SelectionNotSpecified()
                        elif part not in pos:
                            raise SelectionOutOfRange()

                        if rest == '__key__':
                            return part

                        pos = pos[part]
                    else:
                        if rest != None and '*' in rest and '_selection' not in rest:
                            raise SelectionNotSpecified()
                        else:
                            return None
                elif isinstance(pos,list):
                    if part == '*':
                        part = PathHelper.getValue(selection, selectionPath+'_selection')
                    else:
                        try:
                            part = int(part)
                        except:
                            raise SelectionNotSpecified()

                    if part == None:
                        raise SelectionNotSpecified()
                    elif part >= len(pos) or part == -1:
                        raise SelectionOutOfRange()

                    pos = pos[part]
                elif rest != None and '*' in rest:
                    raise SelectionNotSpecified()
                else:
                    return None

            if asList == True and not isinstance(pos, list):
                #if isinstance(pos, dict):
                #    keys = sorted(pos)
                #    values = []
                #    for key in keys:
                #        values.append(pos[key])
                #    pos = values
                #else:
                #    pos = [pos]

                if wrap:
                    pos = [pos]
                else:
                    return None

            if asDict == True and not isinstance(pos, dict):
                #if isinstance(pos, list):
                #    values = {}
                #    for i in range(len(pos)):
                #        values[i] = pos[i]
                #    pos = values
                #else:
                #    pos = {"_undef_":pos}
                if wrap:
                    pos = {"_undef_":pos}
                else:
                    return None

            return pos
        return None

    @staticmethod
    def setValue(base, path, value, selection=None, merge=False, isSub=False, recurse=False):
        if base == None: return

        if path != None and re.search("__root__", path):
            path = path.replace("__root__.", "").replace("__root__", "")

        if value != None and (merge or recurse):
            if isinstance(value, dict):
                for k, v in value.items():
                    if not isSub and k != None and len(k) > 0 and k[0] == '_': continue 
                    subpath = k
                    if path != None: subpath = path + "." + subpath 

                    PathHelper.setValue(base, subpath, v, selection, merge=merge, isSub=True, recurse=recurse)

                return
            elif isinstance(value, (list, tuple)):
                for k in range(len(value)):
                    v = value[k]

                    subpath = unicode(k)
                    if path != None: subpath = path + "." + subpath

                    PathHelper.setValue(base, subpath, v, selection, merge=merge, isSub=True, recurse=recurse)

                return

        pos = base
        lastpos = None
        part = None
        lastpart = None
        selectionPath = ""

        parts = str(path).split('.')
        if len(parts) > 0:
            for i in range(len(parts)):
                lastpart = part
                part = parts[i]
                selectionPath = selectionPath+part+'.'

                if part == '*' and selection != None:
                    part = PathHelper.getValue(selection, selectionPath+'_selection')

                    if part == None:
                        raise SelectionNotSpecified()

                    if isinstance(pos,dict):
                        if part not in pos:
                            raise SelectionOutOfRange()
                    else:
                        if pos == None:
                            lastpos[lastpart] = [] 
                            pos = lastpos[lastpart]

                        elif part >= len(pos) or part == -1:
                            raise SelectionOutOfRange()

                        elif not isinstance(pos,list):
                            lastpos[lastpart] = [] 
                            pos = lastpos[lastpart]
                elif not isinstance(pos, (dict, list, tuple) ):
                    lastpos[lastpart] = {}
                    pos = lastpos[lastpart]

                if isinstance(pos, (list, tuple)):
                    part = int(part)

                if i == len(parts)-1:
                    if not merge or (part not in pos or pos[part] == None):
                        pos[part] = value
                    elif value != None and merge and (part in pos):
                        if isinstance(pos[part], (str, unicode)) and isinstance(value, (float, int)):
                            try:
                                pos[part] = int(pos[part])
                            except:
                                try:
                                    pos[part] = float(pos[part])
                                except:
                                    pos["_old_" + part] = pos[part]
                                    pos[part] = value
                        
                else:
                    if isinstance(pos, dict) and part not in pos:
                        pos[part] = None
                    
                    lastpos = pos
                    pos = pos[part]
